/********************************************************************
 * wanec_iface.c   WANPIPE Echo Canceller Layer (WANEC)
 *
 *
 *
 * ==================================================================
 *
 * May	10	2006	Alex Feldman
 *			Initial Version
 *
 * January 9	2008	David Rokhvarg
 *			Added support for Sangoma MS Windows Driver
 *
 ********************************************************************/

/*=============================================================
 * Includes
 */
#undef WAN_DEBUG_FUNC

#include "wanpipe_includes.h"
#include "wanpipe_defines.h"
#include "wanpipe_debug.h"
#include "wanpipe.h"
//#include "wanpipe_ec_kernel.h"

#if defined(__FreeBSD__) || defined(__OpenBSD__)
# include <sdla_cdev.h>
#elif defined(__LINUX__)
# include "if_wanpipe.h"
# include "wanpipe_tdm_api.h"
#elif defined(__WINDOWS__)
# include "wanpipe_tdm_api.h"
#endif

#include "wanec_iface_api.h"

/*=============================================================
 * Definitions
 */


/*=============================================================
 * Global Parameters
 */
static wan_echocan_iface_t wanec_iface_mc;

extern int wanec_verbose;

/*=============================================================
 * Function definition
 */

extern int wanec_fe2ec_channel(wan_ec_dev_t*, int);

/*****************************************************************************/

static int wanec_mc_channel_opmode_modify(wan_ec_dev_t *ec_dev, int fe_chan, UINT32 opmode, int verbose)
{
	wan_ec_t	*ec = NULL;

	WAN_ASSERT(ec_dev == NULL);
	WAN_ASSERT(ec_dev->ec == NULL);
	WAN_ASSERT(ec_dev->card == NULL);
	ec = ec_dev->ec;

	if (ec_dev->state != WAN_EC_DEV_STATE_READY){
		DEBUG_EVENT(
		"ERROR: %s: Invalid Echo Canceller %s API state (%s)\n",
				ec_dev->devname,
				ec->name,
				WAN_EC_DEV_STATE_DECODE(ec_dev->state));
		return -EINVAL;
	}

	switch(opmode){
	case cOCT6100_ECHO_OP_MODE_NORMAL:
	case cOCT6100_ECHO_OP_MODE_POWER_DOWN:
		break;
	default:
		DEBUG_EVENT(
		"%s: Invalid Echo Channel Operation Mode (opmode=%X)\n",
				ec_dev->devname, opmode);
		return -EINVAL;
	}

	/* Enable Echo cancelation on Oct6100 */
	PRINT1(verbose,
	"%s: Modify Echo Channel OpMode %s on fe_chan:%d ...\n",
			ec_dev->devname,
			(opmode == cOCT6100_ECHO_OP_MODE_NORMAL) ? "Normal" :
			(opmode == cOCT6100_ECHO_OP_MODE_POWER_DOWN) ? "Power Down" : "Unknown",
			fe_chan);
	return wanec_hwctrl(ec_dev, (opmode==cOCT6100_ECHO_OP_MODE_NORMAL)?1:0, fe_chan, verbose);
}


/******************************************************************************
**		WANPIPE EC API INTERFACE FUNCTION
******************************************************************************/

static int wanec_api_modify(wan_ec_dev_t *ec_dev, wan_ec_api_t *ec_api)
{
	wan_ec_t	*ec = NULL;
	u_int32_t	cmd = ec_api->cmd;
	int		fe_chan = 0, err = 0;

	WAN_ASSERT(ec_dev == NULL);
	WAN_ASSERT(ec_dev->ec == NULL);
	ec = ec_dev->ec;

	if (ec_dev->state != WAN_EC_DEV_STATE_READY){
		DEBUG_EVENT(
		"ERROR: %s: Invalid Echo Canceller %s API state (%s)\n",
				ec_dev->devname,
				ec->name,
				WAN_EC_DEV_STATE_DECODE(ec_dev->state));
		return WAN_EC_API_RC_INVALID_STATE;
	}

	if (ec_api->fe_chan_map == 0xFFFFFFFF){
		/* All channels selected */
		ec_api->fe_chan_map = ec_dev->fe_channel_map;
	}

	/* Enable Echo cancelation on Oct6100 */
	PRINT1(ec_api->verbose,
	"%s: %s Echo Canceller on channel(s) chan_map=0x%08lX ...\n",
			ec_dev->devname,
			(cmd == WAN_EC_API_CMD_ENABLE) ? "Enable" : "Disable",
			ec_api->fe_chan_map);
	/*for(chan = fe_first; chan <= fe_last; chan++){*/
	for(fe_chan = ec_dev->fe_start_chan; fe_chan <= ec_dev->fe_stop_chan; fe_chan++){
		if (!(ec_api->fe_chan_map & (1 << fe_chan))){
			continue;
		}
		if (ec_dev->fe_media == WAN_MEDIA_E1 && fe_chan == 0) continue;
		if (cmd == WAN_EC_API_CMD_ENABLE){
			err = wanec_hwctrl(ec_dev, 1, fe_chan, ec_api->verbose);
		}else{
			err = wanec_hwctrl(ec_dev, 0, fe_chan, ec_api->verbose);
		}
		if (err){
			return WAN_EC_API_RC_FAILED;
		}
	}

	return 0;
}

static int wanec_api_modify_hwec(wan_ec_dev_t *ec_dev, wan_ec_api_t *ec_api)
{
	wan_ec_t	*ec = NULL;
	unsigned int	fe_chan = 0;

	WAN_ASSERT(ec_dev == NULL);
	WAN_ASSERT(ec_dev->ec == NULL);
	ec = ec_dev->ec;

	if (ec_dev->state != WAN_EC_DEV_STATE_READY){
		DEBUG_EVENT(
		"ERROR: %s: Invalid Echo Canceller %s API state (%s)\n",
				ec_dev->devname,
				ec->name,
				WAN_EC_DEV_STATE_DECODE(ec_dev->state));
		return WAN_EC_API_RC_INVALID_STATE;
	}
	if (ec_api->fe_chan_map == 0xFFFFFFFF){
		/* All channels selected */
		ec_api->fe_chan_map = ec_dev->fe_channel_map;
	}
	/* Enable/Disable hwec on Oct6100 */
	PRINT1(ec_api->verbose,
	"%s: %s hw echo canceller on channel(s) chan_map=0x%08lX ...\n",
			ec_dev->devname,
			(ec_api->cmd == WAN_EC_API_CMD_HWEC_ENABLE) ? "Enable" : "Disable",
			ec_api->fe_chan_map);
	for(fe_chan = ec_dev->fe_start_chan; fe_chan <= (unsigned int)ec_dev->fe_stop_chan; fe_chan++){
		if (!(ec_api->fe_chan_map & (1 << fe_chan))){
			continue;
		}
		if (ec_api->cmd == WAN_EC_API_CMD_HWEC_ENABLE){
			/* Change rx/tx traffic through Oct6100 */
			if (wanec_hwctrl(ec_dev, 1, fe_chan, ec_api->verbose)){
				return WAN_EC_API_RC_FAILED;
			}
		}else{
			/* Change rx/tx traffic through Oct6100 */
			if (wanec_hwctrl(ec_dev, 0, fe_chan, ec_api->verbose)){
				return WAN_EC_API_RC_FAILED;
			}
		}
	}
	return 0;
}

static int wanec_api_chan_opmode(wan_ec_dev_t *ec_dev, wan_ec_api_t *ec_api)
{
	wan_ec_t	*ec = NULL;
	int		fe_chan, err;

	WAN_ASSERT(ec_dev == NULL);
	WAN_ASSERT(ec_dev->ec == NULL);
	ec = ec_dev->ec;
	if (ec_dev->state != WAN_EC_DEV_STATE_READY){
		DEBUG_EVENT(
		"ERROR: %s: Invalid Echo Canceller %s API state (%s)\n",
				ec_dev->devname,
				ec->name,
				WAN_EC_DEV_STATE_DECODE(ec_dev->state));
		return WAN_EC_API_RC_INVALID_STATE;
	}

	if (ec_api->fe_chan_map == 0xFFFFFFFF){
		/* All channels selected */
		ec_api->fe_chan_map = ec_dev->fe_channel_map;
	}
	/* Enable/Disable Normal mode on Oct6100 */
	PRINT1(ec_api->verbose,
	"%s: Modify Echo Canceller opmode on channel(s) chan_map=0x%08lX ...\n",
			ec_dev->devname, ec_api->fe_chan_map);
	for(fe_chan = ec_dev->fe_start_chan; fe_chan <= ec_dev->fe_stop_chan; fe_chan++){
		if (!(ec_api->fe_chan_map & (1 << fe_chan))){
			continue;
		}
		if (ec_dev->fe_media == WAN_MEDIA_E1 && fe_chan == 0) continue;
		err = wanec_mc_channel_opmode_modify(ec_dev, fe_chan, ec_api->u_chan_opmode.opmode, ec_api->verbose);
		if (err){
			return WAN_EC_API_RC_FAILED;
		}
	}
	return 0;
}

int wanec_mc_ioctl(wan_ec_dev_t *ec_dev, wan_ec_api_t *ec_api)
{
	wan_ec_t	*ec = NULL;
	int		err = 0;
	wan_smp_flag_t	flags;

	WAN_DEBUG_FUNC_START;

	WAN_ASSERT(ec_dev == NULL);
	WAN_ASSERT(ec_dev->ec == NULL);
	ec = ec_dev->ec;

	ec_api->err = WAN_EC_API_RC_OK;
	if (ec_api->cmd == WAN_EC_API_CMD_GETINFO){
		ec_api->u_info.max_channels	= ec->max_ec_chans;
		ec_api->chip_state		= ec->state;
		ec_api->dev_state		= ec_dev->state;
		goto wanec_mc_ioctl_exit;
	}else if (ec_api->cmd == WAN_EC_API_CMD_CONFIG_POLL){
		ec_api->chip_state		= ec->state;
		ec_api->dev_state		= ec_dev->state;
		goto wanec_mc_ioctl_exit;
	}

	wan_spin_lock(&ec->lock,&flags);
	if (wan_test_bit(WAN_EC_BIT_CRIT_DOWN, &ec_dev->critical)){
		DEBUG_EVENT(
		"%s: Echo Canceller device is down!\n",
				ec_api->devname);
		ec_api->err = WAN_EC_API_RC_INVALID_DEV;
		goto wanec_mc_ioctl_done;
	}
	if (wan_test_and_set_bit(WAN_EC_BIT_CRIT_CMD, &ec->critical)){
		DEBUG_EVENT(
		"%s: Echo Canceller is busy!\n",
				ec_api->devname);
		ec_api->err = WAN_EC_API_RC_BUSY;
		goto wanec_mc_ioctl_done;
	}
	switch(ec_api->cmd){
	case WAN_EC_API_CMD_GETINFO:
		ec_api->u_info.max_channels	= ec->max_ec_chans;
		ec_api->chip_state		= ec->state;
		ec_api->dev_state		= ec_dev->state;
		break;
	case WAN_EC_API_CMD_CONFIG_POLL:
		ec_api->chip_state		= ec->state;
		ec_api->dev_state		= ec_dev->state;
		break;
	case WAN_EC_API_CMD_CONFIG:
		ec->state = WAN_EC_CHIP_STATE_OPEN;
		ec_dev->state = WAN_EC_DEV_STATE_READY;
		break;
	case WAN_EC_API_CMD_HWEC_ENABLE:
	case WAN_EC_API_CMD_HWEC_DISABLE:
		err = wanec_api_modify_hwec(ec_dev, ec_api);
		break;
	case WAN_EC_API_CMD_ENABLE:
	case WAN_EC_API_CMD_DISABLE:
		err = wanec_api_modify(ec_dev, ec_api);
		break;
	case WAN_EC_API_CMD_OPMODE:
		err = wanec_api_chan_opmode(ec_dev, ec_api);
		break;
	case WAN_EC_API_CMD_RELEASE:
		ec->state = WAN_EC_CHIP_STATE_RESET;
		ec_dev->state = WAN_EC_DEV_STATE_RESET;
		break;
	case WAN_EC_API_CMD_CHANNEL_OPEN:
	case WAN_EC_API_CMD_CHANNEL_MUTE:
	case WAN_EC_API_CMD_CHANNEL_UNMUTE:
	case WAN_EC_API_CMD_MODIFY_CHANNEL:
	case WAN_EC_API_CMD_TONE_ENABLE:
	case WAN_EC_API_CMD_TONE_DISABLE:
	case WAN_EC_API_CMD_STATS:
	case WAN_EC_API_CMD_STATS_FULL:
	case WAN_EC_API_CMD_STATS_IMAGE:
	case WAN_EC_API_CMD_MONITOR:
		PRINT2(ec_api->verbose,
		"%s: Unsupported Echo Canceller ioctl command %s!\n",
				ec_api->devname, WAN_EC_API_CMD_DECODE(ec_api->cmd));
		err = 0;
		break;
	case WAN_EC_API_CMD_RELEASE_ALL:
		break;
	}
	if (err){
		PRINT2(ec_api->verbose,
		"%s: %s return error (Command: %s)\n",
			ec_api->devname, __FUNCTION__, WAN_EC_API_CMD_DECODE(ec_api->cmd));
		ec_api->err = err;
	}
	if (ec_api->err == WAN_EC_API_RC_INVALID_STATE){
		ec_api->chip_state	= ec->state;
		ec_api->dev_state	= ec_dev->state;
	}
	wan_clear_bit(WAN_EC_BIT_CRIT_CMD, &ec->critical);

wanec_mc_ioctl_done:
	wan_spin_unlock(&ec->lock,&flags);

wanec_mc_ioctl_exit:
	WAN_DEBUG_FUNC_END;
	return 0;
}


/*****************************************************************************/
static void*
wanec_mc_register(void *pcard, u_int32_t fe_port_mask, int max_line_no, int max_channels, void *pconf)
{
	sdla_t			*card = (sdla_t*)pcard;
	wan_ec_dev_t		*ec_dev = NULL;

	WAN_DEBUG_FUNC_START;
	WAN_ASSERT_RC(pcard == NULL, NULL);
	ec_dev = wanec_register(WAN_ECHOCAN_ID_MC, pcard, fe_port_mask, max_line_no, max_channels);
	if (ec_dev == NULL){
		DEBUG_EVENT("%s: ERROR: Failed to create Echo Canceller interface!\n",
					card->devname);
		return NULL;
	}
	return (void*)ec_dev;
}

static int wanec_mc_unregister(void *arg, void *pcard)
{
	sdla_t		*card = (sdla_t*)pcard;
	wan_ec_t	*ec = NULL;
	wan_ec_dev_t	*ec_dev = (wan_ec_dev_t*)arg;
	wan_smp_flag_t	flags;

	WAN_DEBUG_FUNC_START;
	WAN_ASSERT(ec_dev == NULL);
	WAN_ASSERT(ec_dev->ec == NULL);
	ec = ec_dev->ec;

	wan_spin_lock(&ec->lock,&flags);

	wan_set_bit(WAN_EC_BIT_TIMER_KILL,(void*)&ec_dev->critical);
	wan_set_bit(WAN_EC_BIT_CRIT_DOWN,(void*)&ec_dev->critical);
	wan_set_bit(WAN_EC_BIT_CRIT,(void*)&ec_dev->critical);
	wan_del_timer(&ec_dev->timer);

	wan_spin_unlock(&ec->lock,&flags);

	if (wanec_unregister(ec_dev, pcard)){
		DEBUG_EVENT("%s: Failed to unregister Echo Canceller device %s!\n",
					card->devname, ec->name);
		return -EINVAL;
	}

	WAN_DEBUG_FUNC_END;
	return 0;
}

static int wanec_event_ctrl(void *arg, wan_event_ctrl_t *ectrl)
{
	wan_ec_t	*ec = NULL;
	wan_ec_dev_t	*ec_dev = (wan_ec_dev_t*)arg;
	int		err = 0;
	wan_smp_flag_t	flags;

	WAN_ASSERT(ec_dev == NULL);
	WAN_ASSERT(ec_dev->ec == NULL);
	WAN_ASSERT(ectrl  == NULL);
	ec = ec_dev->ec;

	wan_spin_lock(&ec->lock, &flags);
	if (wan_test_and_set_bit(WAN_EC_BIT_CRIT_CMD, &ec->critical)){
		wan_spin_unlock(&ec->lock, &flags);
		return -EBUSY;
	}

	switch(ectrl->type){
	case WAN_EVENT_HWEC:
		if (ectrl->mode == WAN_EVENT_ENABLE){
			wanec_hwctrl(ec_dev, 1, ectrl->channel, 0);
		}else{
			wanec_hwctrl(ec_dev, 0, ectrl->channel, 0);
		}
		break;
	default:
		err = -EINVAL;
		break;
	}
	wan_clear_bit(WAN_EC_BIT_CRIT_CMD, &ec->critical);
	wan_spin_unlock(&ec->lock, &flags);
	return err;
}


int wanec_mc_init(void)
{
	memset(&wanec_iface_mc, 0, sizeof(wan_echocan_iface_t));
#if defined(CONFIG_WANPIPE_HWEC)
	wanec_iface_mc.id	= WAN_ECHOCAN_ID_MC,
	wanec_iface_mc.reg	= wanec_mc_register,
	wanec_iface_mc.unreg	= wanec_mc_unregister,
	wanec_iface_mc.isr	= NULL,
	wanec_iface_mc.poll	= NULL,
	wanec_iface_mc.ectrl	= wanec_event_ctrl,
#endif
	register_wanec_iface(&wanec_iface_mc, wanec_verbose);
	return 0;
}

int wanec_mc_release(void)
{
	unregister_wanec_iface(&wanec_iface_mc, wanec_verbose);
	memset(&wanec_iface_mc, 0, sizeof(wan_echocan_iface_t));
	return 0;
}
